import six
from abc import ABCMeta, abstractmethod
from eyewitness.config import BBOX


@six.add_metaclass(ABCMeta)
class ObjectClassifier():
    """
    Abstract class used to wrapper object detector
    """
    @abstractmethod
    def detect(self, image_obj, bbox_objs=None):
        """
        [abstract method] need to implement detection method which return DetectionResult obj

        Parameters
        ----------
        image_obj: eyewitness.image_util.Image

        bbox_objs: Optional[List[BoundedBoxObject]]

        Returns
        -------
        DetectionResult: DetectionResult
            the detected result of given image

            if bbox_objs not given:
                -> classify whole image, and result in DETECTED_OBJECTS (length 1)
            if bbox_objs given:
                -> classify List of bbox_objs, and return valid label objs in DetectionResult

        # TODO: need to consider a more proper design
        """
        pass

    @property
    def detection_method(self):
        """
        detection_method for the ObjectDetector is BBOX

        Returns
        -------
        detection_method: String
        """
        return BBOX

    @property
    def valid_labels(self):
        """
        [abstract property] the valid_labels of this detector e.g. set(['person', 'pikachu' ...])
        this will be used while want to evaluation the detector

        Returns
        -------
        valid_labels: set[String]
        """
        raise NotImplementedError



from collections import namedtuple

# detection type
BBOX = 'bbox'

# Feedback obj type
FEEDBACK_BBOX_OBJ = 'feedback_bbox'
FEEDBACK_NO_OBJ = 'feedback_no_obj'

# image producer type
POST_PATH = 'post_path'
POST_BYTES = 'post_bytes'
IN_MEMORY = 'in_memory'

# Line false_alert msg template
LINE_FALSE_ALERT_MSG_TEMPLATE = "false_alert_{image_id}_{meta}"
LINE_FALSE_ALERT_ANNOTATION_PATTERN = (
    r"false_alert_(?P<channel>.*)--(?P<timestamp>\d*)--(?P<file_format>.*)_(?P<meta>.*)")

# audience_platfrom
LINE_PLATFROM = 'line'

# image_info fields
RAW_IMAGE_PATH = 'raw_image_path'
IMAGE_ID = 'image_id'
IMAGE_ID_STR_TEMPLATE = '{channel}--{timestamp}--{file_format}'
IMAGE_ID_STR_PATTERN = r'(?P<channel>.*)--(?P<timestamp>\d*)--(?P<file_format>.*)'

# detected_result fields
DRAWN_IMAGE_PATH = 'drawn_image_path'
DETECTED_OBJECTS = 'detected_objects'
DETECTION_METHOD = 'detection_method'

# audience_info fields
AUDIENCE_ID = 'audience_id'
RECEIVE_TIME = 'receive_time'
AUDIENCE_ID_STR_TEMPLATE = '{platform_id}--{user_id}'
AUDIENCE_ID_STR_PATTERN = r'(?P<platform_id>.*)--(?P<user_id>.*)'

# feedback infomation
FEEDBACK_METHOD = 'feedback_method'
FEEDBACK_MSG_OBJS = 'feedback_msg_objs'
FEEDBACK_META = 'feedback_meta'
IS_FALSE_ALERT = 'is_false_alert'


# detected object types:
BoundedBoxObject = namedtuple(
    'BoundedBoxObject', ['x1', 'y1', 'x2', 'y2', 'label', 'score', 'meta'])

DETECTED_OBJECT_TYPE_MAPPING = {BBOX: BoundedBoxObject}


# feedback object types:
BboxObjectFeedback = namedtuple('BboxObjectFeedback', ['x1', 'y1', 'x2', 'y2', 'label', 'meta'])
FEEDBACK_OBJECT_TYPE_MAPPING = {
    FEEDBACK_BBOX_OBJ: BboxObjectFeedback,
    FEEDBACK_NO_OBJ: None,  # this kind of feedback
}

# dataset config
DATASET_TRAIN_AND_VALID = 'TRAIN_AND_VALID'
DATASET_TEST_ONLY = 'TEST_ONLY'
DATASET_ALL = 'ALL'



from abc import ABCMeta, abstractmethod
from collections import defaultdict

from eyewitness.config import (
    BBOX,
    DATASET_TEST_ONLY,
)
from eyewitness.utils import bbox_intersection_over_union
import logging
import six
import numpy as np

LOG = logging.getLogger(__name__)


@six.add_metaclass(ABCMeta)
class Evaluator():
    @property
    def evaluation_method(self):
        raise NotImplementedError

    def evaluate(self, detector, dataset):
        # do the type check.
        assert all(_type == self.evaluation_method
                   for _type in [detector.detection_method, dataset.dataset_type])
        return self._evaluate(detector, dataset)

    @abstractmethod
    def _evaluate(self, detector, dataset):
        pass


class BboxMAPEvaluator(Evaluator):
    """
    evaluate the bbox mAP score
    """
    def __init__(self, iou_threshold=0.5, dataset_mode=DATASET_TEST_ONLY, logging_frequency=100):
        self.iou_threshold = iou_threshold
        self.dataset_mode = dataset_mode
        self.logging_frequency = logging_frequency

    @property
    def evaluation_method(self):
        return BBOX

    def _evaluate(self, detector, dataset):
        """
        Parameters
        ----------
        detector: ObjectDetector
            object detector instance

        dataset: BboxDataSet
            BboxDataSet instance

        Returns
        -------
        eval_result: Dict
            - mAP: mAP score
            - ap_stats: ap_stats
        """
        # only calculate on overlapped labels
        valid_labels = detector.valid_labels.intersection(dataset.valid_labels)
        if not valid_labels:
            raise ValueError('no valid labels can used for evaluation')

        selected_images = dataset.get_selected_images(mode=self.dataset_mode)
        n_images = len(selected_images)

        # forward all images, and collect all ground truth from dataset
        detected_objs = defaultdict(list)
        gt_objs = defaultdict(lambda: defaultdict(list))
        gt_label_count = defaultdict(int)
        processed_image_count = 0
        for image_obj, image_gt_objs in dataset.dataset_iterator(mode=self.dataset_mode):
            processed_image_count += 1
            if processed_image_count % self.logging_frequency == 0:
                LOG.info('processed images %d/%d', processed_image_count, n_images)

            detection_result = detector.detect(image_obj)
            for obj in detection_result.detected_objects:
                if obj.label not in valid_labels:
                    continue
                detected_objs[obj.label].append((image_obj.image_id, obj))

            # stores gt information into gt_objs and gt_label_count
            image_gt_objs = [image_gt_obj for image_gt_obj in
                             image_gt_objs if image_gt_obj.label in valid_labels]
            for image_gt_obj in image_gt_objs:
                gt_objs[image_obj.image_id][image_gt_obj.label].append(image_gt_obj)
                gt_label_count[image_gt_obj.label] += 1

        ap_stats = self.calculate_label_ap(valid_labels, detected_objs, gt_objs, gt_label_count)
        mAP = np.mean(list(i['AP'] for i in ap_stats.values()))
        return {'mAP': mAP, 'ap_stats': ap_stats}

    def calculate_label_ap(self, valid_labels, detected_objs, gt_objs, gt_label_count):
        """
        refactor the evaluation from
        https://github.com/rafaelpadilla/Object-Detection-Metrics
        """
        ap_stat = {}

        for label in valid_labels:  # loop each label
            n_positives = gt_label_count[label]
            sorted_label_det_objs = sorted(
                detected_objs[label], key=lambda x: x[1].score, reverse=True)
            TP = np.zeros(len(sorted_label_det_objs))
            FP = np.zeros(len(sorted_label_det_objs))

            for det_idx, (image_id, label_obj) in enumerate(sorted_label_det_objs):
                # find label gt truth image obj
                if image_id not in gt_objs:
                    gt_label_objs = []
                else:
                    gt_label_objs = gt_objs[image_id][label]

                # get the max_iou
                best_iou = 0
                best_iou_gt_idx = -1
                for gt_idx, gt_label_obj in enumerate(gt_label_objs):
                    det_bbox = label_obj[:4]  # x1, y1, x2, y2
                    gt_bbox = gt_label_obj[:4]  # x1, y1, x2, y2
                    iou = bbox_intersection_over_union(det_bbox, gt_bbox)
                    if iou > best_iou:
                        best_iou_gt_idx = gt_idx
                        best_iou = iou

                if best_iou >= self.iou_threshold:
                    TP[det_idx] = 1
                    # remove the matched gt obj from this image
                    gt_objs[image_id][label] = (gt_label_objs[:best_iou_gt_idx]
                                                + gt_label_objs[(best_iou_gt_idx + 1):])
                else:
                    FP[det_idx] = 1

            acc_FP = np.cumsum(FP)
            acc_TP = np.cumsum(TP)
            recall = acc_TP / n_positives
            precision = np.divide(acc_TP, (acc_FP + acc_TP))
            ap, mpre, mrec, _ = BboxMAPEvaluator.calculate_average_precision(recall, precision)

            ap_stat[label] = {
                'precision': precision,
                'recall': recall,
                'AP': ap,
                'interpolated precision': mpre,
                'interpolated recall': mrec,
                'total positives': n_positives,
                'total TP': np.sum(TP),
                'total FP': np.sum(FP)
            }
            LOG.info('ap calculated for %s: %s', label, ap_stat[label]['AP'])
        return ap_stat

    @staticmethod
    def calculate_average_precision(recall, precision):
        mrec = []
        mrec.append(0)
        [mrec.append(e) for e in recall]
        mrec.append(1)
        mpre = []
        mpre.append(0)
        [mpre.append(e) for e in precision]
        mpre.append(0)
        for i in range(len(mpre) - 1, 0, -1):
            mpre[i - 1] = max(mpre[i - 1], mpre[i])
        ii = []
        for i in range(len(mrec) - 1):
            if mrec[1:][i] != mrec[0:-1][i]:
                ii.append(i + 1)
        ap = 0
        for i in ii:
            ap = ap + np.sum((mrec[i] - mrec[i - 1]) * mpre[i])
        # return [ap, mpre[1:len(mpre)-1], mrec[1:len(mpre)-1], ii]
        return [ap, mpre[0:len(mpre) - 1], mrec[0:len(mpre) - 1], ii]



import six
import logging
from abc import ABCMeta, abstractmethod
from collections import defaultdict

import arrow
from eyewitness.models.db_proxy import DATABASE_PROXY
from eyewitness.models.detection_models import BboxDetectionResult
from eyewitness.models.feedback_models import FalseAlertFeedback
from eyewitness.config import (
    BBOX,
    BoundedBoxObject,
    DETECTED_OBJECTS
)

from eyewitness.utils import bbox_intersection_over_union

LOG = logging.getLogger(__name__)


@six.add_metaclass(ABCMeta)
class DetectionResultFilter():
    @property
    def detection_method(self):
        raise NotImplementedError

    def apply(self, detection_result):
        assert self.detection_method == detection_result.detection_method
        return self._apply(detection_result)

    @abstractmethod
    def _apply(self, detection_result):
        """
        Parameters
        ----------
        detection_result: DetectionResult

        Retruns
        -------
        filtered_detection_result: DetectionResult
        """


class FeedbackBboxDeNoiseFilter(DetectionResultFilter):
    def __init__(self, database, decay=0.9, iou_threshold=0.7, collect_feedback_period=172800,
                 detection_threshold=0.5, time_check_period=None):
        """
        a Bbox DeNoise filter, which will read false alert bbox from tables:
        FalseAlertFeedback, BboxDetectionResult, and apply filter onto the detection result
        """
        self.database = database
        self.decay = decay
        self.iou_threshold = iou_threshold
        self.collect_feedback_period = collect_feedback_period
        if time_check_period is None:
            self.time_check_period = collect_feedback_period * 0.005
        else:
            self.time_check_period = time_check_period

        DATABASE_PROXY.initialize(self.database)
        # the FalseAlertFeedback, BboxDetectionResult table must exist in same database
        FalseAlertFeedback.create_table()
        BboxDetectionResult.create_table()

        # get the false alert feedbacks
        self.collect_feedback_water_mark = None
        self.update_false_alert_feedback_bbox()
        self.detection_threshold = detection_threshold

    @property
    def detection_method(self):
        """
        detection_method: String
            BBOX
        """
        return BBOX

    def _apply(self, detection_result):
        """
        using the false-alert feedback msgs to de-noise the detection result

        loop over false-alert objects related to (channel, classes),
        if any detected object matched false alert object with iou > iou_threshold:
            `object_score *= decay`
        """
        # try to collect latest false alert feedback
        self.update_false_alert_feedback_bbox()

        # verify if all detected objs having enough confidence.
        passed_objs = []
        channel = detection_result.image_id.channel
        detected_objs = detection_result.image_dict.get(DETECTED_OBJECTS, [])
        for detected_obj in detected_objs:
            (x1, y1, x2, y2, label, score, meta) = detected_obj
            bbox = (x1, y1, x2, y2)

            # loop over false_alert obj, decay the matched obj score
            false_alert_objs = self.false_alert_feedback_bbox.get((channel, label), [])
            for feedback_obj_bbox in false_alert_objs:
                if bbox_intersection_over_union(bbox, feedback_obj_bbox) > self.iou_threshold:
                    score *= self.decay
                    # TODO: meta should be a more flexible way
                    meta += 'decay_%s |' % (self.decay)

            if score > self.detection_threshold:
                passed_objs.append(BoundedBoxObject(x1, y1, x2, y2, label, score, meta))
        LOG.info('original: %s objs, filtered: %s objs', len(detected_objs), len(passed_objs))
        detection_result.image_dict[DETECTED_OBJECTS] = passed_objs
        return detection_result

    def update_false_alert_feedback_bbox(self):
        """
        collect_bbox_false_alert_information
        """
        self.check_proxy_db()

        now_timestamp = arrow.now().timestamp

        if self.collect_feedback_water_mark is None:
            do_update = True
        else:
            time_check = self.collect_feedback_water_mark + self.time_check_period
            do_update = now_timestamp > time_check

        if do_update:
            LOG.info('collecting latest false alert feedback')
            self.collect_feedback_water_mark = arrow.now().timestamp
            start_time = now_timestamp - self.collect_feedback_period
            query = FalseAlertFeedback.select().where(FalseAlertFeedback.receive_time > start_time)
            feedback_dict = defaultdict(list)
            for feedback in query:
                image_id = feedback.image_id
                channel = image_id.channel
                detected_objs = list(
                    BboxDetectionResult.select().where(BboxDetectionResult.image_id == image_id))
                for obj in detected_objs:
                    label = obj.label
                    bbox = (obj.x1, obj.y1, obj.x2, obj.y2)
                    feedback_dict[(channel, label)].append(bbox)

            self.false_alert_feedback_bbox = feedback_dict
            for key, values in feedback_dict.items():
                LOG.info('collecting %s objects from %s', len(values), key)

    def check_proxy_db(self):
        """check if the db proxy is correct one, if not initialize again.
        """
        if not (self.database is DATABASE_PROXY.obj):
            DATABASE_PROXY.initialize(self.database)
